//=====================================================================
// PMON_LZHUF.cpp :
//
// pmon_parser()
// fbb_parser()
//
// buffer handling routines
//
// Written by Hans-Peter Helfert (DL6MAA)
// (C) 2019-2020 SCS GmbH & Co. KG, Hanau, Germany
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//
//=====================================================================

#include <errno.h>
#include <fcntl.h> 
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <ctype.h>

#define MAXTOTALBYTES 100000

char payload2array[MAXTOTALBYTES - 1];   // array to store the ASCII CHARS ("02,FA,BG....") of all payload 2's
char outputarray[MAXTOTALBYTES / 3];
extern FILE* infile;

#define BUFLEN 256
#define BINBUFLEN 100000

char linebuffer[BUFLEN];
unsigned char binbuffer[BINBUFLEN];

int inptr = 0;
int outptr = 0;

int binptr = 0;
int boutptr = 0;

char testbuf[256];
int parser_state = 0;
unsigned int type;
int len;
int charcount;

int fbb_state = 0;
int uncompressed_len = 0;
unsigned char logical_len = 0;

int lzhuf_decoder_otf(unsigned char c, FILE *);

void reset_lzhuf_decoding(void)
{
	extern int lzhuf_decoder_state;

	fbb_state = 0;
	parser_state = 0;
	lzhuf_decoder_state = 0;
}

/*************************************************************
put_buffer()

*************************************************************/

void put_buffer(char c)
{
	linebuffer[inptr] = c;
	inptr++;
	inptr %= BUFLEN;
}

/*************************************************************
get_buffer()

*************************************************************/

int get_buffer(char* c)
{
	if (inptr == outptr)
		return 0;

	*c = linebuffer[outptr];
	outptr++;
	outptr %= BUFLEN;

	return 1;
}

/*************************************************************
retrieve_buffer()

*************************************************************/

void retrieve_buffer(char* bufx, int len)
{
	int i;
	int outptrx;

	outptrx = inptr + BUFLEN - len;
	outptrx %= BUFLEN;

	if (len == 0)
		return;

	for (i = 0; i < len; i++)
	{
		bufx[i] = linebuffer[outptrx];
		outptrx++;
		outptrx %= BUFLEN;
	}
	bufx[i] = 0;
}

/*************************************************************
put_binbuffer()

*************************************************************/

void put_binbuffer(unsigned char c)
{
	binbuffer[binptr] = c;
	binptr++;
	binptr %= BINBUFLEN;
}

/*************************************************************
get_buffer()

*************************************************************/

int get_binbuffer(unsigned char* c)
{
	if (binptr == boutptr)
		return 0;

	*c = binbuffer[boutptr];
	boutptr++;
	boutptr %= BINBUFLEN;

	return 1;
}

/*************************************************************
retrieve_buffer()

*************************************************************/

void retrieve_binbuffer(unsigned char* bufx, int len)
{
	int i;
	int outptrx;

	if (len == 0)
		return;

	outptrx = binptr + BINBUFLEN - len;
	outptrx %= BINBUFLEN;

	for (i = 0; i < len; i++)
	{
		bufx[i] = binbuffer[outptrx];
		outptrx++;
		outptrx %= BINBUFLEN;
	}
	bufx[i] = 0;
}

/*************************************************************
rewind_buffer()

*************************************************************/

void rewind_binbuffer(unsigned char* bufx, int len)
{
	int i;
	int outptrx;

	if (len == 0)
		return;

	outptrx = boutptr + BINBUFLEN - len;
	outptrx %= BINBUFLEN;

	for (i = 0; i < len; i++)
	{
		bufx[i] = binbuffer[outptrx];
		outptrx++;
		outptrx %= BINBUFLEN;
	}
	bufx[i] = 0;
}


/*************************************************************
fbb_parser()

*************************************************************/

void fbb_parser(unsigned char l, FILE *outfile)
{
	int i;
	unsigned char bufferx[30];
	//unsigned char c;
	extern int lzhuf_decoder_state;
	int numbytes;

	switch (fbb_state)
	{
	case 0:
		rewind_binbuffer(bufferx, 25);

		if (strncmp("FC EM", (char*)bufferx, 5) == 0)
		{
			printf("\n\n###====== <FC EM> HEADER FOUND, STARTING LZHUF DECOMPRESSION ======###\n");
			fprintf(outfile, "\n\n###====== <FC EM> HEADER FOUND, STARTING LZHUF DECOMPRESSION ======###\n");

			for (i = 6; i < 24; i++)
			{
				if (bufferx[i] == ' ') // look for next space character
				{
					uncompressed_len = atoi((char*)bufferx + i);
					break;
				}
			}

			if (uncompressed_len)
			{
				printf("### UNCOMPRESSED LEN OF MESSAGE: %d\n", uncompressed_len);
			}

			fbb_state++;
		}
		break;

	case 1:
		rewind_binbuffer(bufferx, 3);
		if ((bufferx[0] + bufferx[2] * 65536 + bufferx[1] * 256) == 12288)  // 0x00, 0x03, 0x00 sequence
		{
			printf("\n### START OF COMPRESSED MESSAGE FOUND\n");
			lzhuf_decoder_state = 0;				 // reset "on the fly" lzhuf-decoder
			fbb_state++;
		}
		break;

	case 2:
		fbb_state++;
		break;

	case 3:
		rewind_binbuffer(bufferx, 2);

		if (bufferx[0] == 2)
		{
			logical_len = (unsigned char) bufferx[1];
			printf("\n\n### LOGICAL PACKET FOUND, LEN %d\n", logical_len);
			printf("\n### LZHUF OUTPUT (on the fly):\n\n");
		
			fbb_state++;
		}
		else
		{
			if (bufferx[0] == 4)
			{
				printf("\n\n###====== MESSAGE COMPLETED, DECOMPRESSION FINISHED ======###\n\n");
				fprintf(outfile, "\n\n###====== MESSAGE COMPLETED, DECOMPRESSION FINISHED ======###\n\n");
				//	printf("\ndebugcode: %d, %d\n", bufferx[0], bufferx[1]);
			}
			else
			{
				printf("\n###====== STOPPING LZHUF DECOMPRESSION: LOGICAL PACKET NOT FOUND! ======###\n");
				fprintf(outfile, "\n###====== STOPPING LZHUF DECOMPRESSION: LOGICAL PACKET NOT FOUND! ======###\n");
				//	printf("\ndebugcode: %d, %d\n", bufferx[0], bufferx[1]);
			}
			fbb_state = 0;
		}
		break;

	case 4:
		//printf("\nLogical Len: %d\n", logical_len);

		if (logical_len > 0)
		{
			logical_len--;
			numbytes = lzhuf_decoder_otf(l, outfile);

			//printf("Bytes: %d\n", numbytes);

			if (numbytes >= uncompressed_len - 2)
			{
				reset_lzhuf_decoding();
				printf("\n\n###====== MESSAGE COMPLETED, DECOMPRESSION FINISHED ======###\n\n");
				fprintf(outfile, "\n\n###====== MESSAGE COMPLETED, DECOMPRESSION FINISHED ======###\n\n");
				return;
			}

			if (logical_len == 0)
				fbb_state = 2;
		}
		break;
	}
}

/*************************************************************
pmon_parser()

*************************************************************/

int pmon_parserx(char c, FILE *outfile)
{
	int i;
	static int hexcnt;
	static int printcnt;

	//printf("%c", c);
	put_buffer(c);

	switch (parser_state)
	{
	case 0:
		type = 0;
		len = 0;
		parser_state++;
		break;

	case 1:
		retrieve_buffer(testbuf, 12);

		if (strncmp("LEN: ", testbuf, 5) == 0)
		{
			//printf("\n\nFOUND LEN!!!\n\n");
			len = atoi(testbuf + 5);

			//printf("\nLEN: %d\n", len);
		}

		if (strncmp("TYPE: ", testbuf, 6) == 0)
		{
			//printf("\n\nFOUND TYPE!!!\n\n");
			type = atoi(testbuf + 6);

			//printf("\nType: %d\n", type);
		}

		if (strncmp("###PAYLOAD2:", testbuf, 12) == 0)
		{
			//printf("\n\nFOUND PAYLOAD2:\n");
			if (type == 8)
			{
				len = 3 * len - 1;	//Hex numbers + comma per byte (except at the end of line)
				printf("\n### <Bin data>\n");
			}
			else
				len += 2;

			parser_state++;
			charcount = 0;
			hexcnt = 0;
		}
		break;

	case 2:
		//if (type == 8)
		{
			// get rid of "newline" characters...
			if (c == 10)
				break;
			if (c == 13)
				break;
		}

		parser_state++;

	case 3:

#define TBDELAY0 15		// must not be smaller than 2 because auf auto-linefeed (below)
		retrieve_buffer(testbuf, TBDELAY0);

		if (len > 0)
		{
			if (type < 8)
			{
			//	printcnt++;
			//	if (printcnt < 3)
			//		break;

				if ((len < 3) && ((c == '#') || (c == 13) || (c == 10)))
				{
					//TODO, check "###PAYLOAD_END" "on the fly" and suppress output 
					parser_state = 0;
					break;
				}
				//else
				{
					// auto linefeed on console 
				//	if ((testbuf[TBDELAY0-2] == 13) && (c != 13))
				//		printf("%c", 10); // insert LF

					if (c == 13)
						printf("%c", 10);

					printf("%c", c); // copy printable payload to console

					fflush(stdout);
					
					fputc(c, outfile);
					put_binbuffer((unsigned char)c);
				}
			}
			else
			{
				hexcnt++;
				if (hexcnt == 2)
				{
					retrieve_buffer(testbuf, 2);
				
					i = 0;
					testbuf[2] = 0;
					sscanf_s(testbuf, "%x", &i);
					//printf("%s,", testbuf);
					//printf("%x,", i & 0xFF);
					put_binbuffer(i & 0xFF);
				}

				if (hexcnt == 3)
					hexcnt = 0;
			}

			len--;
		}
		else
			parser_state = 0;

		charcount++;

		if (len <= 0)
		{
			parser_state = 0;	
		}

		// stop parsing for sure if we find this string
		if (strncmp("###PAYLOAD_END", testbuf, 14) == 0)
			parser_state = 0;

		break;
	}

	return 0;
}
