//============================================================================
// PMON_lzh.cpp : project "main" file
//
// Entry point of program after start
// Single threaded loop() with 100 ms timeslots
//
// Written by Hans-Peter Helfert (DL6MAA)
// (C) 2019-2020 SCS GmbH & Co. KG, Hanau, Germany
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//
//============================================================================

#include <iostream>
#include <stdint.h>
#include <string.h>
#include <errno.h>
#include <conio.h>
#include "rs232.h"

//#define DEBUGX

FILE* infile = NULL, * outfile = NULL, *outfile1 = NULL;

int timeout = 0;
#define FBB_TIMEOUT 300  // 60 seconds

int main(int argc, char* argv[])
{
	errno_t err;
	//int errx;
	char c;

	extern int fbb_state; // TEST TEST TEST

	int pmon_parser(FILE *in);
	int Decode(FILE *out);

	int pmon_parserx(char, FILE*);
	void fbb_parser(unsigned char l, FILE*);

	int get_binbuffer(unsigned char*);

	char filename[256];
	//char* fname;
	char comname[256];

	unsigned int i, n;
	int cport_nr = 1;        
	int	bdrate = 115200;       

	unsigned char buf[4096];
	char mode[] = { '8','N','1',0 };
	int state = 0;
	char command_string[256];
	unsigned int pactor_level = 3;

	//int len;
	int parser_state = 0;
	unsigned char l;

	printf("\nPMON LZHUF-Decoder Version 1.0.7 (24-SEP-2019)\n");
	printf(  "==============================================\n\n");

	if (argv[1] == NULL)
	{
		printf("PMON_lzh: ERROR:\n");
		printf("Missing input file, please append filename of capture file as first argument...\n");
		printf("Or append COM port number, e.g. COM25, for continuous monitoring using an SCS P4dragon modem...\n");
		printf("EXAMPLE: pmon_lzh COM25<Enter>\n");
		printf("Look for COM port numbers in Windows Device Manager...\n");
		exit(-1);
	}

	for (i = 0; i < strlen(argv[1]); i++)
		comname[i] = toupper(argv[1][i]);

	comname[i] = 0;

	if (strncmp("COM", comname, 3) == 0)
	{
		cport_nr = atoi(comname + 3) - 1;

		printf("\nTrying to open COM port %d...\n", cport_nr + 1);

		if (RS232_OpenComport(cport_nr, bdrate, mode, 0))
		{
			printf("Can not open comport\n");

			exit(-5);
		}

		// open log file
		err = fopen_s(&outfile, "pmon_payload_data.txt", "a");

		if (err != 0)
		{
			printf("Unable to open file : pmon_payload_data.txt\n");
			exit(-3);
		}

		// open log file
		err = fopen_s(&outfile1, "pmon_original_data.txt", "a");

		if (err != 0)
		{
			printf("Unable to open file : pmon_original_data.txt\n");
			exit(-4);
		}


		if (outfile != NULL)
			fprintf(outfile, "\n\n==========================================================================================\n\n");


		if (outfile1 != NULL)
			fprintf(outfile1, "\n\n==========================================================================================\n\n");

		printf("\nNOW FETCHING DATA FROM MODEM, press any key to abort\n");
	    printf(  "====================================================\n\n");

		while (1)
		{
			switch (state)
			{
			case 0:
				//strcpy_s(command_string, 256, "\'VER\r");
				//RS232_SendBuf(cport_nr, (unsigned char*)command_string, 5);
				state++;
				Sleep(100);
				fetch_data(cport_nr);
				break;

			case 1:
				if (argv[2] != NULL)	// read second command line argument, if available
				{
					i =  (unsigned int) atoi(argv[2]);
					if ((i == 3) || (i == 4))
						pactor_level = i;
				}

				//strcpy_s(command_string, 256, "pmon mode 3\r");
				sprintf_s(command_string, "\rpmon mode %d\r\0", pactor_level);
				RS232_SendBuf(cport_nr, (unsigned char*)command_string, 13);
				state++;
				Sleep(100);
				fetch_data(cport_nr);
				break;

			case 2:
#ifdef DEBUGX
				strcpy_s(command_string, 256, "pmon V 0\r");  // verbos 0 is only for testing
				RS232_SendBuf(cport_nr, (unsigned char*)command_string, 9);

				strcpy_s(command_string, 256, "pmon H 0\r");  // Hex 0  is only for testing
				RS232_SendBuf(cport_nr, (unsigned char*)command_string, 9);
#else
				strcpy_s(command_string, 256, "pmon V 1\r");  // verbos 0 is only for testing
				RS232_SendBuf(cport_nr, (unsigned char*)command_string, 9);

				strcpy_s(command_string, 256, "pmon H 1\r");  // Hex 0  is only for testing
				RS232_SendBuf(cport_nr, (unsigned char*)command_string, 9);
#endif

				strcpy_s(command_string, 256, "pmon P 0\r");  // Packet 0  
				RS232_SendBuf(cport_nr, (unsigned char*)command_string, 9);

				Sleep(100);
				fetch_data(cport_nr);
				state++;
				break;

			case 3:
				strcpy_s(command_string, 256, "pmon Start\r");
				RS232_SendBuf(cport_nr, (unsigned char*)command_string, 11);
				Sleep(100);
				fetch_data(cport_nr);
				state++;
				break;

			case 4:
				printf("\n\n### LISTENING TO PMON...\n\n");
				state++;
				break;

			case 5:
				n = RS232_PollComport(cport_nr, buf, 4095);

				for (i = 0; i < n; i++)
				{
					if (outfile1 != NULL)
						fprintf(outfile1, "%c", buf[i]);

					parser_state = pmon_parserx(buf[i], outfile);
				}

				//printf("\nFBB_STATE: %d\n", fbb_state);

				while (get_binbuffer(&l))
				{
					//printf("%c", l);

					// parsing FBB data (header and logical packets)
					fbb_parser(l, outfile);
					timeout = 0;
				}

				Sleep(100);
				timeout++;

				if (timeout > FBB_TIMEOUT)
				{
					if (fbb_state)
					{
						fbb_state = 0;
						printf("\n###===== STOPPING LZHUF DECOMPRESSION: *** TIMEOUT *** =====###\n");
					}
				}
				break;
			}

#ifdef _WIN32
			Sleep(100);

			if (_kbhit())
			{
				printf("\nPMON OFF");

				strcpy_s(command_string, 256, "pmon Stop\r");
				RS232_SendBuf(cport_nr, (unsigned char*)command_string, 11);

				Sleep(100);

				printf("\n\nTHANK YOU FOR USING PMON_lzh!\n");

				// close logfile
				if (outfile != NULL)
					fclose(outfile); 

				exit(0);
			}
#else
			usleep(100000);  /* sleep for 100 milliSeconds */
#endif
		}
	}

	err = fopen_s(&infile, argv[1], "r");

	if (err != 0)
	{
		printf("Unable to open file: %s\n", argv[1]);
		exit(-1);
	}

	strcpy_s(filename, "decompressed_");
	strcat_s(filename, argv[1]);

	err = fopen_s(&outfile, filename, "wb");

	if (err != 0)
	{
		if(infile != NULL)
			fclose(infile);

		printf("Unable to open file : %s\n", filename);
		exit(-3);
	}
	
	parser_state = 0;

	// parse PMON packets and collect payload in circular buffer
	do
	{
		if (infile != NULL)
		{
			c = fgetc(infile);
			pmon_parserx(c, outfile);
		}
		else
			break;

	} while (c != EOF);

	fbb_state = 0;

	// FFB parse collected payload
	while (get_binbuffer(&l))
	{
		//printf("%c", l);

		// parsing FBB data (header and logical packets)
		fbb_parser(l, outfile);
	}

	if (infile != NULL)
		fclose(infile);
	if (outfile != NULL)
		fclose(outfile);
	if (outfile1 != NULL)
		fclose(outfile1);
}
